#include <tamtypes.h>
#include <libcdvd.h>
#include <debug.h>
#include <fileio.h>
#include <iopcontrol.h>
#include <iopheap.h>
#include <kernel.h>
#include <loadfile.h>
#include <sbv_patches.h>
#include <sifrpc.h>


static int resetiop = 0;

extern u32 usbd_irx;
extern u32 usb_mass_irx;
extern u32 cdvdinit_irx;

extern u32 size_usbd_irx;
extern u32 size_usb_mass_irx;
extern u32 size_cdvdinit_irx;

void (*unpack)(u32 src, u32 dest) = (void*)0x008004D0;

void ResetIOP()
{
	if(resetiop)
	{
		fioExit();
		SifExitIopHeap();
		SifLoadFileExit();
		SifExitRpc();
	}

	SifIopReset("rom0:UDNL rom0:EELOADCNF", 0);
    
	while ( SifIopSync()) ;

	resetiop = 1;

	SifInitRpc(0);
}


void LoadElf(const char *elf, char* path)
{
	char* args[1];
	t_ExecData exec;	

	SifLoadElf(elf, &exec);

	#ifdef DEBUG
		scr_printf("Trying to load ELF: %s\n", elf);
	#endif	

	if(exec.epc > 0)
	{	
		FlushCache(0);
		FlushCache(2);

		// Reset IOP, since not all ELF's do it and we've loaded additional IOP
		// modules which need to be unloaded
		ResetIOP();

		if(path != 0)
		{
			args[0] = path;
			ExecPS2((void*)exec.epc, (void*)exec.gp, 1, args);
		}
		else
		{
			ExecPS2((void*)exec.epc, (void*)exec.gp, 0, NULL);
		}
	}
}

#ifdef NTSC

#define OFFSET	0x01000000

void ntsc_patch()
{
	int addr;
	u32 patch_addr = 0;

	// Find SetGsCrt call
	for(addr=OFFSET; addr < OFFSET+(3*1024*1024) && !patch_addr; addr+=4)
	{
		if(( _lw(addr) == 0x24050003) && (_lw(addr+4) == 0x0c059960))
			patch_addr = addr;
	}
	#ifdef DEBUG
		scr_printf("patch_addr: 0x%08X\n", patch_addr);
	#endif

	if(patch_addr)
	{
		#ifdef DEBUG
			scr_printf("Before [0x%08X] = 0x%08X\n", patch_addr, _lw(patch_addr));
		#endif

		_sw(0x00000000, 0x00800988); // nop unpack function
		_sw(0x24050002, patch_addr); // Change SetGsCrt arg 2 from 3 (PAL) to 2 (NTSC)
	
		#ifdef DEBUG
			scr_printf("After  [0x%08X] = 0x%08X\n", patch_addr, _lw(patch_addr));
		#endif
	}
}

#endif


void LoadManager(int port)
{
	t_ExecData exec;	

	if(port == 0) SifLoadElf("mc0:/BESLES-00000 MP/RUNME.ELF", &exec);
	if(port == 1) SifLoadElf("mc1:/BESLES-00000 MP/RUNME.ELF", &exec);

	if(exec.epc > 0)
	{	
		FlushCache(0);
		FlushCache(2);

		#ifdef NTSC
		if( _lw(0x00800988) == 0x0c200134) // Check that we are patching the correct ELF
		{
			// Call the code-unpacking-function
			unpack(0x00800A00, 0x01000000);
			// Once the code is unpacked, patch to NTSC
			ntsc_patch();

			FlushCache(0);
			FlushCache(2);
		}
		#endif

		ExecPS2((void*)exec.epc, (void*)exec.gp, 0, NULL);
	}
}


int GetPort()
{
	int fd;

	if( (fd = fioOpen("mc0:/BESLES-00000 MP/RUNME.ELF", O_RDONLY)) > 0)
	{
		fioClose(fd);
		return 0;
	}

	#ifdef DEBUG
		scr_printf("Couldn't open mc0:/BESLES-00000 MP/RUNME.ELF\n");
	#endif	

	if( (fd = fioOpen("mc1:/BESLES-00000 MP/RUNME.ELF", O_RDONLY)) > 0)
	{
		fioClose(fd);
		return 1;
	}
	
	#ifdef DEBUG
		scr_printf("Couldn't open mc1:/BESLES-00000 MP/RUNME.ELF\n");
	#endif	

	DEBUG_BGCOLOR(0xFF);
	ExitDeleteThread();
	
	return -1;
}

int _start()
{
	int ret;
	int port;

	#ifdef DEBUG
		init_scr();
	#endif

	ResetIOP();

	// Allow loading of ELF/IRX from memory card
	sbv_patch_disable_prefix_check();	

	SifLoadFileInit();

	// Load CD/DVD initialization module
	SifExecModuleBuffer(&cdvdinit_irx, size_cdvdinit_irx, 0, NULL, &ret);
	
	// Reset IOP and stop CD/DVD drive	
	ResetIOP();

	#ifdef DEBUG
		scr_printf("cdInit..");
	#endif

	cdInit(CDVD_INIT_INIT);
	cdStop();
	cdSync(0);

	#ifdef DEBUG
		scr_printf("ok!\n");
	#endif

	// Load Memory Card modules from BIOS
	SifLoadModule("rom0:SIO2MAN", 0, NULL);
	SifLoadModule("rom0:MCMAN", 0, NULL);
	SifLoadModule("rom0:MCSERV", 0, NULL);

	
	// Find which port the Memory Plus memory card is in
	port = GetPort();

	#ifdef DEBUG
		scr_printf("Memory Card Port %i\n", port);
	#endif

	// Try to load from memory card
	if(port == 0) LoadElf("mc0:/BESLES-00000 MP/EXECUTE1.ELF", "mc0:/BESLES-00000 MP/");
	if(port == 1) LoadElf("mc1:/BESLES-00000 MP/EXECUTE2.ELF", "mc1:/BESLES-00000 MP/");
	
	#ifdef DEBUG
		scr_printf("Loading from memory card failed, trying USB mass\n");
	#endif

	// Unable to load from memory card, try USB pendrive
	
	// Load USB drivers
	SifExecModuleBuffer(&usbd_irx, size_usbd_irx, 0, NULL, &ret);
	SifExecModuleBuffer(&usb_mass_irx, size_usb_mass_irx, 0, NULL, &ret);

	if(port == 0) LoadElf("mass:EXECUTE1.ELF", "mass:");
	if(port == 1) LoadElf("mass:EXECUTE2.ELF", "mass:");
	
	#ifdef DEBUG
		scr_printf("Loading from USB mass failed, loading Memory Plus manager\n");
	#endif

	// EXECUTE1/2.ELF not found, load memory card manager
	LoadManager(port);
	
	return 0;
}
